@@ -105,7 +105,8 @@ class BASE_SERIAL_DEVICE:
         self.output = None
         self.wr_cmd = self.cmd
         self.prompt = b'>>> '
-        self.dev_description, self.manufacturer, self._hwid = self._get_serial_port_data(serial_port)
+        self.dev_description, self.manufacturer, self._hwid = self._get_serial_port_data(
+            serial_port)
         self.serial = serial.Serial(serial_port, baudrate)
 
     def _get_serial_port_data(self, serialport):
@@ -141,9 +142,11 @@ class BASE_SERIAL_DEVICE:
         if self._traceback in self.buff:
             long_string = True
         if long_string:
-            self.response = self.buff.replace(b'\r', b'').replace(b'\r\n>>> ', b'').replace(b'>>> ', b'').decode()
+            self.response = self.buff.replace(b'\r', b'').replace(
+                b'\r\n>>> ', b'').replace(b'>>> ', b'').decode()
         else:
-            self.response = self.buff.replace(b'\r\n', b'').replace(b'\r\n>>> ', b'').replace(b'>>> ', b'').decode()
+            self.response = self.buff.replace(b'\r\n', b'').replace(
+                b'\r\n>>> ', b'').replace(b'>>> ', b'').decode()
         if not silent:
             if self.response != '\n' and self.response != '':
                 print(self.response)
@@ -166,18 +169,24 @@ class BASE_SERIAL_DEVICE:
         if not hr:
             self.bytes_sent = self.serial.write(bytes(self._reset, 'utf-8'))
         else:
-            self.bytes_sent = self.serial.write(bytes(self._hreset, 'utf-8'))
+            try:
+                self.bytes_sent = self.serial.write(bytes(self._hreset, 'utf-8'))
+            except OSError:
+                pass
         time.sleep(0.5)
-        self.buff = self.serial.read_all()
+        try:
+            self.buff = self.serial.read_all()
+        except OSError:
+            pass
         if not silent:
             print('Done!')
 
-    def kbi(self, silent=True, pipe=None):
+    def kbi(self, silent=True, pipe=None, long_string=False):
         if pipe is not None:
             self.wr_cmd(self._kbi, silent=silent)
             pipe(self.response, std='stderr')
         else:
-            self.cmd(self._kbi, silent=silent)
+            self.cmd(self._kbi, silent=silent, long_string=long_string)
 
     def banner(self, pipe=None):
         self.cmd(self._banner, silent=True, long_string=True)
@@ -248,8 +257,8 @@ class SERIAL_DEVICE(BASE_SERIAL_DEVICE):
         os.uname().machine, unique_id(), sys.implementation.name]"
         (self.dev_platform, self._release,
          self._version, self._machine, uuid, imp) = self.cmd(repr_cmd,
-                                                  silent=True,
-                                                  rtn_resp=True)
+                                                             silent=True,
+                                                             rtn_resp=True)
         vals = hexlify(uuid).decode()
         imp = imp[0].upper() + imp[1:]
         imp = imp.replace('p', 'P')
@@ -258,13 +267,13 @@ class SERIAL_DEVICE(BASE_SERIAL_DEVICE):
         dev_str = '(MAC: {})'.format(self._mac)
         desc_str = '{}, Manufacturer: {}'.format(self.dev_description,
                                                  self.manufacturer)
-        return 'SerialDevice @ {}, Type: {}, \
-Class: {}\nFirmware: {}\n{}\n{}'.format(self.serial_port,
-                                        self.dev_platform,
-                                        self.dev_class,
-                                        fw_str,
-                                        desc_str,
-                                        dev_str)
+        return (f'SerialDevice @ {self.serial_port}, Type: {self.dev_platform}, '
+                f'Class: {self.dev_class}\n'
+                f'Firmware: {fw_str}\n{desc_str}\n{dev_str}')
+
+    @property
+    def address(self):
+        return self.serial_port
 
     def flush_conn(self):
         flushed = 0
@@ -325,7 +334,8 @@ Class: {}\nFirmware: {}\n{}\n{}'.format(self.serial_port,
                     self.paste_cmd = ''
                     if pipe is None:
                         print('')  # print Traceback under ^C
-                        self.kbi(pipe=pipe, silent=False)  # KBI
+                        self.kbi(pipe=pipe, silent=False,
+                                 long_string=long_string)  # KBI
                     else:
                         self.kbi(pipe=pipe)  # KBI
                     time.sleep(0.2)
@@ -335,13 +345,16 @@ Class: {}\nFirmware: {}\n{}\n{}'.format(self.serial_port,
         cmd_filt = bytes(cmd + '\r\n', 'utf-8')
         self.buff = self.buff.replace(cmd_filt, b'', 1)
         if dlog:
-            self.data_buff = self.buff.replace(b'\r', b'').replace(b'\r\n>>> ', b'').replace(b'>>> ', b'').decode()
+            self.data_buff = self.buff.replace(b'\r', b'').replace(
+                b'\r\n>>> ', b'').replace(b'>>> ', b'').decode()
         if self._traceback in self.buff:
             long_string = True
         if long_string:
-            self.response = self.buff.replace(b'\r', b'').replace(b'\r\n>>> ', b'').replace(b'>>> ', b'').decode()
+            self.response = self.buff.replace(b'\r', b'').replace(
+                b'\r\n>>> ', b'').replace(b'>>> ', b'').decode()
         else:
-            self.response = self.buff.replace(b'\r\n', b'').replace(b'\r\n>>> ', b'').replace(b'>>> ', b'').decode()
+            self.response = self.buff.replace(b'\r\n', b'').replace(
+                b'\r\n>>> ', b'').replace(b'>>> ', b'').decode()
         if not silent:
             if self.response != '\n' and self.response != '':
                 if pipe is None:
@@ -396,18 +409,22 @@ Class: {}\nFirmware: {}\n{}\n{}'.format(self.serial_port,
                     if self.prompt in self.message:
                         break
             else:
-                self.message = self.serial.read_all()
+                self.message = b''
+                while b'\r\n' not in self.message:
+                    self.message += self.serial.read(1)
+                    if self.prompt in self.message:
+                        break
             self.buff += self.message
             self.raw_buff += self.message
             if self.message == b'':
                 pass
             else:
-                if self.message.startswith(b'\n'):
+                if self.message.startswith(b'\n') and 'ls(' not in inp:
                     self.message = self.message[1:]
                 if pipe:
                     cmd_filt = bytes(inp + '\r\n', 'utf-8')
                     self.message = self.message.replace(cmd_filt, b'', 1)
-                msg = self.message.replace(b'\r', b'').decode()
+                msg = self.message.replace(b'\r', b'').decode('utf-8')
                 if 'cat' in inp:
                     if msg.endswith('>>> '):
                         msg = msg.replace('>>> ', '')
@@ -427,10 +444,12 @@ Class: {}\nFirmware: {}\n{}\n{}'.format(self.serial_port,
                                         if self._traceback.decode() in pipe_out:
                                             self._is_traceback = True
                                             # catch before traceback:
-                                            pipe_stdout = pipe_out.split(self._traceback.decode())[0]
+                                            pipe_stdout = pipe_out.split(
+                                                self._traceback.decode())[0]
                                             if pipe_stdout != '' and pipe_stdout != '\n':
                                                 pipe(pipe_stdout)
-                                            pipe_out = self._traceback.decode() + pipe_out.split(self._traceback.decode())[1]
+                                            pipe_out = self._traceback.decode(
+                                            ) + pipe_out.split(self._traceback.decode())[1]
                                         if self._is_traceback:
                                             pipe(pipe_out, std='stderr')
                                         else:
@@ -446,10 +465,12 @@ Class: {}\nFirmware: {}\n{}\n{}'.format(self.serial_port,
                                     if self._traceback.decode() in pipe_out:
                                         self._is_traceback = True
                                         # catch before traceback:
-                                        pipe_stdout = pipe_out.split(self._traceback.decode())[0]
+                                        pipe_stdout = pipe_out.split(
+                                            self._traceback.decode())[0]
                                         if pipe_stdout != '' and pipe_stdout != '\n':
                                             pipe(pipe_stdout)
-                                        pipe_out = self._traceback.decode() + pipe_out.split(self._traceback.decode())[1]
+                                        pipe_out = self._traceback.decode(
+                                        ) + pipe_out.split(self._traceback.decode())[1]
                                     if self._is_traceback:
                                         pipe(pipe_out, std='stderr')
                                     else:
@@ -465,10 +486,12 @@ Class: {}\nFirmware: {}\n{}\n{}'.format(self.serial_port,
                                 if self._traceback.decode() in pipe_out:
                                     self._is_traceback = True
                                     # catch before traceback:
-                                    pipe_stdout = pipe_out.split(self._traceback.decode())[0]
+                                    pipe_stdout = pipe_out.split(
+                                        self._traceback.decode())[0]
                                     if pipe_stdout != '' and pipe_stdout != '\n':
                                         pipe(pipe_stdout)
-                                    pipe_out = self._traceback.decode() + pipe_out.split(self._traceback.decode())[1]
+                                    pipe_out = self._traceback.decode(
+                                    ) + pipe_out.split(self._traceback.decode())[1]
                                 if self._is_traceback:
                                     pipe(pipe_out, std='stderr')
                                 else:
@@ -544,7 +567,8 @@ Class: {}\nFirmware: {}\n{}\n{}'.format(self.serial_port,
                 fs = int((1/time_out)*1000)
             if fs is not None:
                 temp_dict['fs'] = fs
-                temp_dict['ts'] = [i/temp_dict['fs'] for i in range(len(temp_dict[dvars[0]]))]
+                temp_dict['ts'] = [i/temp_dict['fs']
+                                   for i in range(len(temp_dict[dvars[0]]))]
             if units is not None:
                 temp_dict['u'] = units
             self.datalog = temp_dict
