@@ -26,7 +26,7 @@ from .serialdevice import *
 from .websocketdevice import *
 from .devgroup import *
 from .decorators import *
-from .bledevice import *
+# from .bledevice import *
 from .exceptions import *
 from ipaddress import ip_address
 import socket
@@ -36,11 +36,14 @@ def check_device_type(dev_address, resolve_name=False):
     if isinstance(dev_address, str):
         if '.' in dev_address and dev_address.count('.') == 3:
             # check IP
-            try:
-                ip_address(dev_address)
-                return 'WebSocketDevice'
-            except Exception as e:
-                print(e)
+            if ':' not in dev_address:
+                try:
+                    ip_address(dev_address)
+                    return 'WebSocketDevice'
+                except Exception as e:
+                    print(e)
+            else:
+                return check_device_type(dev_address.split(':')[0], resolve_name)
         elif dev_address.endswith('.local'):
             try:
                 if resolve_name:
@@ -49,6 +52,8 @@ def check_device_type(dev_address, resolve_name=False):
                     return 'WebSocketDevice'
             except Exception as e:
                 print(e)
+        elif '.local' in dev_address and ':' in dev_address:
+            return check_device_type(dev_address.split(':')[0], resolve_name)
         elif 'COM' in dev_address or '/dev/' in dev_address:
             return 'SerialDevice'
         elif len(dev_address.split('-')) == 5:
@@ -78,10 +83,9 @@ def Device(dev_address, password=None, **kargs):
             baudrt = fkargs.pop('baudrate')
         return SerialDevice(dev_address, baudrate=baudrt, **fkargs)
     if dev_type == 'WebSocketDevice':
-        if dev_address.endswith('.local'):
-            dev_address = socket.gethostbyname(dev_address)
         return WebSocketDevice(dev_address, password, **kargs)
     if dev_type == 'BleDevice':
+        from .bledevice import BleDevice
         pop_args = ['ssl', 'auth', 'capath']
         fkargs = {k: v for k, v in kargs.items() if k not in pop_args}
         return BleDevice(dev_address, **fkargs)
